#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include <rox/options.h>
#include <rox/rox_filer_action.h>
#include "menu_images.h"
#include "roxmenu.h"
#include "utils.h"

typedef enum {
	MIS_APPDIR,
	MIS_REGFILE,
	MIS_DESKTOPFILE
} MenuItemSourceType;

static gchar *get_label_text(const gchar *path);
static gint dir_sort(RoxMenuDir *a, RoxMenuDir *b);
static gint file_sort(RoxMenuItem *a, RoxMenuItem *b);
static gint path_sort(gchar *a, gchar *b);
static gint menudir_path_is_equal(RoxMenuDir *a, gchar *b);
static void menudir_make_widgets(RoxMenuDir *dir);
static void menudir_update_dirlist(RoxMenuDir *dir, GList *dirlist);
static void menudir_update_filelist(RoxMenuDir *dir, GList *filelist);
static void menudir_update_contents(RoxMenuDir *dir);
static void menudir_add_subdirs_to_menu(RoxMenuDir *dir);
static void menudir_add_files_to_menu(RoxMenuDir *dir);
static GtkWidget *menudir_create_menuitem(RoxMenuDir *dir);
static void menudir_update_modtime(RoxMenuDir *dir);
static gboolean menudir_has_changed(RoxMenuDir *mdir);

/* Utility function prototypes. */
static gboolean include_path(const char *path);
static gboolean include_path_dir(RoxMenuDir *dir, const char *path);
static int run_path(GtkWidget *w, gpointer filepath);
/*static int run_appdir(const char *dirpath);*/

/* Utility functions. */

/* Get the menu item label, stripping prefix if necessary. */
static gchar *get_label_text(const gchar *path) {

	extern Option strip_number;
	
	gchar *ret, *base;
	char *start_str;
	
	base = g_path_get_basename(path);
	ret = base;
	
	if ( strip_number.int_value && isdigit(base[0]) > 0 ) {
		start_str = strchr(base, '_');
		start_str = start_str != NULL ? (start_str + 1) : base;
		ret = g_strdup(start_str);
		g_free(base);
	} 
	return ret;
}

static gint dir_sort(RoxMenuDir *a, RoxMenuDir *b) {
	gint ret;
	ret = path_sort(a->path, b->path);
	return ret;
}

static gint file_sort(RoxMenuItem *a, RoxMenuItem *b) {
	gint ret;
	ret = path_sort(a->path, b->path);
	return ret;
}

static gint menudir_path_is_equal(RoxMenuDir *a, gchar *b) {
	gint ret;
	ret = strcmp(a->path, b);
	return ret;
}

/* Takes two paths, a and b, and sorts by case-insensitive basename. */
static gint path_sort(gchar *a, gchar *b) {
	gint ret;
	gchar *s1=NULL, *s2=NULL, *tmp;
	tmp = g_path_get_basename(a);
	s1 = g_utf8_casefold(tmp, -1);
	g_free(tmp);
	tmp = g_path_get_basename(b);
	s2 = g_utf8_casefold(tmp, -1);
	g_free(tmp);
	
	ret = g_utf8_collate(s1, s2);
	g_free(s1);
	g_free(s2);
	return ret;
}

static gboolean include_path(const char *path) {
	gchar *base, *ignore;
	gboolean ret = TRUE;
	
	g_return_val_if_fail(path != NULL, FALSE);
	
	if (! g_file_test(path, G_FILE_TEST_EXISTS)) return FALSE;
	
	base = g_path_get_basename(path);
	if (base[0] == '.') ret = FALSE;
	ignore = g_build_path(G_DIR_SEPARATOR_S, path, ".rox-menu_ignore", NULL);
	if (g_file_test(ignore, G_FILE_TEST_EXISTS)) ret = FALSE;
	
	g_free(ignore);
	g_free(base);

	return ret;
}

static gboolean include_path_dir(RoxMenuDir *dir, const char *path) {
	gboolean ret;

	g_return_val_if_fail(dir != NULL && path != NULL, FALSE);
	
	ret = include_path(path);
	if (ret && ! g_file_test(path, G_FILE_TEST_IS_DIR) &&
	           ! is_appdir(path) ) {
		ret = dir->show_files;
	}

	return ret;
}

#if 0
static int run_appdir(const char *dirpath) {
	
	pid_t child;
	gchar *path;
	
	g_return_val_if_fail(is_appdir(dirpath), -1);
	
	path = g_build_path(G_DIR_SEPARATOR_S, dirpath, "AppRun", NULL);	
	child = fork();
	
	if (child == -1) return (0);
	else if (child == 0) {
		char *argv[2];
		argv[0] = path;
		argv[1] = 0;
		execv(path, argv);
		_exit(127);
	}
	
	return child;
	
}
#endif

static int run_path(GtkWidget *w, gpointer filepath) {
	rox_filer_run(filepath);
	return 1;
}

static void menudir_add_subdirs_to_menu(RoxMenuDir *dir) {
	GList *p;
	RoxMenuDir *currdir;
	gint basepos, currpos, newpos;

	g_return_if_fail(dir != NULL);

	if (dir->dir_grouping == GROUPS_ON_BOTTOM) {
		basepos = g_list_length(dir->dir_contents);
		if (dir->dir_contents) basepos = basepos + 1;
	} else {
		basepos = 0;
	}
	
	p = dir->dir_subdirs;
	while (p != NULL) {
		currdir = p->data;
		if (GTK_IS_WIDGET(currdir->visible_item) && 
		    gtk_widget_get_parent(GTK_WIDGET(currdir->visible_item)) == NULL) 
		{
			currpos = g_list_position(dir->dir_subdirs, p);
			newpos = basepos + (currpos < 0 ? 0 : currpos);
			gtk_menu_shell_insert(GTK_MENU_SHELL(dir->visible_menu),
			                      currdir->visible_item, newpos);
		}
		p = g_list_next(p);
	}
}

static void menudir_add_files_to_menu(RoxMenuDir *dir) {
	GList *p;
	RoxMenuItem *currfile;
	gint basepos, currpos, newpos;

	g_return_if_fail(dir != NULL);
	
	if (dir->dir_grouping == GROUPS_ON_TOP) {
		basepos = g_list_length(dir->dir_subdirs);
		if (dir->dir_subdirs) basepos = basepos + 1;
	} else {
		basepos = 0;
	}
	
	p = dir->dir_contents;
	while (p != NULL) {
		currfile = p->data;
		if (GTK_IS_WIDGET(currfile->visible_item) && 
		    gtk_widget_get_parent(GTK_WIDGET(currfile->visible_item)) == NULL)
		{
			currpos = g_list_position(dir->dir_contents, p);
			newpos = basepos + (currpos < 0 ? 0 : currpos);
			gtk_menu_shell_insert(GTK_MENU_SHELL(dir->visible_menu),
			                      currfile->visible_item, newpos);
		}
		p = g_list_next(p);
	}
}

static void menudir_make_widgets(RoxMenuDir *dir) {

	gint sep_pos;

	g_return_if_fail(dir != NULL);

	if (menudir_is_empty(dir)) return;  

	if (dir->dir_grouping == GROUPS_ON_BOTTOM) {
		menudir_add_files_to_menu(dir);
		menudir_add_subdirs_to_menu(dir);
		sep_pos = g_list_length(dir->dir_contents);
	} else {
		menudir_add_subdirs_to_menu(dir);
		menudir_add_files_to_menu(dir);
		sep_pos = g_list_length(dir->dir_subdirs);
	}
	
	if (dir->dir_subdirs != NULL && dir->dir_contents != NULL) {
		if (!dir->sep) {
			dir->sep = gtk_separator_menu_item_new();
			gtk_widget_show(dir->sep);
			gtk_menu_shell_insert(GTK_MENU_SHELL(dir->visible_menu), 
			                      dir->sep, sep_pos);
		}
	} else if(GTK_IS_WIDGET(dir->sep)) {
		gtk_widget_destroy(dir->sep);
		dir->sep = NULL;
	}
	
	if (! dir->toplevel) {
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(dir->visible_item), 
		                          dir->visible_menu);
		gtk_widget_show(dir->visible_item);
	}
	gtk_widget_show(dir->visible_menu);
}

static GtkWidget *menudir_create_menuitem(RoxMenuDir *dir) {
	
	GtkWidget *ret, *pix;
	gchar *name;
	
	g_return_val_if_fail(dir != NULL, NULL);
	
	name = get_label_text(dir->path);
	ret = gtk_image_menu_item_new_with_label(name);
	g_free(name);
	pix = get_menu_pixmap_path(dir->path);
	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(ret), pix);
	g_object_ref(ret); /* Keep the menuitem around until we kill it. */
	gtk_widget_show(GTK_WIDGET(pix));
	gtk_widget_show(ret);
	return ret;
}

static void menudir_update_dirlist(RoxMenuDir *dir, GList *dirlist) {
	
	GList *curr, *p, *rem;
	RoxMenuDir *newdir, *d;
	
	g_return_if_fail(dir != NULL);
	
	/* Remove non-existent entries. */
	p = dir->dir_subdirs;
	while (p != NULL) {
		d = (RoxMenuDir*)(p->data);
		if (! g_file_test(d->path, G_FILE_TEST_IS_DIR) ||
		    ! include_path(d->path) || is_appdir(dir->path) ) {
			rem = p;
			p = g_list_next(p);
			dir->dir_subdirs = g_list_remove_link(dir->dir_subdirs, rem);
			menudir_destroy(rem->data);
			g_list_free_1(rem);
		} 
		else p = g_list_next(p);
	}
	
	/* Add missing items. */
	curr = dirlist;
	while (curr != NULL) {
		p = g_list_find_custom(dir->dir_subdirs, curr->data, 
		                       (GCompareFunc)menudir_path_is_equal);
		if ( p == NULL && include_path(curr->data) ) {
			newdir = menudir_new(curr->data);
			/* Default display properties to the parent. */
			newdir->dir_grouping = dir->dir_grouping;
			newdir->show_files = dir->show_files;
			
			dir->dir_subdirs = g_list_insert_sorted(dir->dir_subdirs,
											newdir, (GCompareFunc)dir_sort);
		}
		curr = g_list_next(curr);
	}
}


static void menudir_update_filelist(RoxMenuDir *dir, GList *filelist) {
	
	GList *curr, *p, *rem;
	RoxMenuItem *newitem,  *d;
	
	g_return_if_fail(dir != NULL);
	
	/* Remove non-existent entries. */
	p = dir->dir_contents;
	while (p != NULL) {
		d = p->data;
		if (! include_path_dir(dir, d->path) ||
		    ( g_file_test(d->path, G_FILE_TEST_IS_DIR) && 
		      ! is_appdir(d->path) ) ){
			rem = p;
			p = g_list_next(p);
			dir->dir_contents = g_list_remove_link(dir->dir_contents, rem);
			menuitem_destroy(rem->data);
			if (rem) g_list_free_1(rem);
			d = NULL;
		}
		else p = g_list_next(p);
	}

	/* Add missing items. */
	curr = filelist;
	while (curr != NULL) {
		p = g_list_find_custom(dir->dir_contents, curr->data, 
		                       (GCompareFunc)menudir_path_is_equal);
		if ( p == NULL && include_path(curr->data) )  {
			newitem = menuitem_new(curr->data);
			dir->dir_contents = g_list_insert_sorted(dir->dir_contents,
											newitem, (GCompareFunc)file_sort);
		}
		curr = g_list_next(curr);
	}
}

static void menudir_update_contents(RoxMenuDir *dir) {
	
	GList *files = NULL, *dirs = NULL, *tmp;
	GDir *curdir;
	G_CONST_RETURN gchar *next_file;
	gchar *full_path;

	g_return_if_fail(dir != NULL);

	/* If the directory hasn't changed, there's no need to update it. */
	if (!menudir_has_changed(dir) ) return;
	
	/* Update the modification time. */
	menudir_update_modtime(dir);

	/* Get a list of the directories and files in mendir. */
	curdir = g_dir_open(dir->path, 0, NULL);
	next_file = g_dir_read_name(curdir);
	
	while (next_file) {
		full_path = g_build_path(G_DIR_SEPARATOR_S, dir->path, next_file, NULL);
		if (include_path(full_path)) { 
			if (is_regdir(full_path)) {
				dirs = g_list_append(dirs, full_path);
			} else if ( is_appdir(full_path) || dir->show_files ) {
				files = g_list_append(files, full_path);
			}
		} else g_free(full_path);
		next_file = g_dir_read_name(curdir);
	}
	
	/* Sort file and directory lists by name. */
	dirs = g_list_sort(dirs, (GCompareFunc)path_sort);
	files = g_list_sort(files, (GCompareFunc)path_sort);
	
	/* Update the file and directory lists. */
	menudir_update_dirlist(dir, dirs);
	menudir_update_filelist(dir, files);
	
	/* Free the lists and associated data. */
	tmp = dirs;
	while (tmp) {
		g_free(tmp->data);
		tmp = g_list_next(tmp);
	}
	g_list_free(dirs);

	tmp = files;
	while (tmp) {
		g_free(tmp->data);
		tmp = g_list_next(tmp);
	}
	g_list_free(files);
}

static gboolean menudir_has_changed(RoxMenuDir *mdir) {
	struct stat fileinfo;
	int status;
	g_return_val_if_fail(mdir != NULL, FALSE);  /* Is this right? */
	status = stat(mdir->path, &fileinfo);
	if (status != 0) return FALSE;   /* Why did I do this...? */
	else return (mdir->modtime != fileinfo.st_mtime);
}

static void menudir_update_modtime(RoxMenuDir *dir) {
	struct stat fileinfo;
	int status;
	g_return_if_fail(dir != NULL);
	status = stat(dir->path, &fileinfo);
	dir->modtime = fileinfo.st_mtime;
}

/************************************/
/******** External Interface ********/
/************************************/

/*** RoxMenuItem object functions ***/

RoxMenuItem *menuitem_new(const char *path) {

	RoxMenuItem *ret = NULL;
	GtkWidget *pixmap;
	gchar *label_text;

	g_return_val_if_fail(g_file_test(path, G_FILE_TEST_EXISTS), NULL);
	
	ret = g_try_malloc(sizeof(RoxMenuItem));
	g_return_val_if_fail(ret != NULL, NULL);

	label_text = get_label_text(path);

	/* Create the menu item, including image and callback. */
	ret->path = g_strdup(path);
	ret->visible_item = gtk_image_menu_item_new_with_label(label_text);
	g_free(label_text);
	/* Keep the widget from being destroyed when the menu is recreated.*/
	g_object_ref(ret->visible_item);
	pixmap = get_menu_pixmap_path((gchar*)path);
	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(ret->visible_item),
	                              pixmap);
	
	/* Not sure if I should depend on the "rox" script for everything. */
	g_signal_connect(GTK_OBJECT(ret->visible_item), "activate", 
			(GtkSignalFunc) run_path, ret->path);
	gtk_widget_show(ret->visible_item);
	
	return ret;

}

void menuitem_destroy(RoxMenuItem *item) {
	g_return_if_fail(item != NULL);
	g_free(item->path);
	gtk_widget_destroy(item->visible_item);
	g_free(item);
}

/*** RoxMenuDir object functions ***/

RoxMenuDir *menudir_new(const char *path) {

	RoxMenuDir *ret;

	g_return_val_if_fail(g_file_test(path, G_FILE_TEST_IS_DIR), NULL);

	/* Make things simple - reuse the toplevel initialization function
	   and just change the few things that need to be changed. */
	ret = menudir_toplevel_new(path);
	g_return_val_if_fail(ret != NULL, NULL);
	
	ret->toplevel = FALSE;
	ret->visible_item = menudir_create_menuitem(ret);
	gtk_widget_show(ret->visible_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(ret->visible_item), 
	                          ret->visible_menu);
	return ret;
}

RoxMenuDir *menudir_toplevel_new(const char *path) {

	RoxMenuDir *ret;

	g_return_val_if_fail(g_file_test(path, G_FILE_TEST_IS_DIR),NULL);

	ret = g_try_malloc(sizeof(RoxMenuDir));
	g_return_val_if_fail(ret != NULL, NULL);

	/* Initialize all the pointers to NULL because I'm paranoid. */
	ret->path = NULL;
	ret->visible_item = NULL;
	ret->visible_menu = NULL;
	ret->sep = NULL;
	ret->dir_subdirs = NULL;
	ret->dir_contents = NULL;
	ret->dir_grouping = GROUPS_ON_TOP;
	ret->is_updating = FALSE;
	ret->show_files = FALSE;
	
	ret->path = g_strdup(path);	
	ret->modtime = (time_t)0;
	ret->visible_menu = gtk_menu_new();
	gtk_widget_show(ret->visible_menu);
	ret->toplevel = TRUE;

	return ret;
}

void menudir_destroy(RoxMenuDir *mendir) {
	GList *rem;
	RoxMenuItem *i;
	RoxMenuDir *d;

	g_return_if_fail(mendir != NULL);
	
	rem = mendir->dir_contents;
	while (rem != NULL) {
		i = rem->data;
		menuitem_destroy(rem->data);
		rem = g_list_next(rem);
	}
	g_list_free(mendir->dir_contents);
	
	rem = mendir->dir_subdirs;
	while (rem != NULL) {
		d = rem->data;
		menudir_destroy(d);
		rem = g_list_next(rem);
	}
	g_list_free(mendir->dir_subdirs);
	
	g_free(mendir->path);
	gtk_widget_destroy(mendir->visible_menu);
	/*if (GTK_IS_WIDGET(mendir->sep)) gtk_widget_destroy(mendir->sep);*/
	if (! mendir->toplevel) gtk_widget_destroy(mendir->visible_item);
}

void menudir_update(RoxMenuDir *dir) {
	GList *l, *rem;
	gboolean changed;
	RoxMenuDir *d;

	g_return_if_fail(dir != NULL);
	
	if (dir->is_updating) return;  /* Don't update during an update. */
	dir->is_updating = TRUE;

	/* This accounts for when the menu updates while copying a big 
	   AppDir.  The AppDir initially looks like a regular directory, but
		becomes an AppDir when the AppRun script is copied.  This checks 
		the directories and makes sure none of them have become AppDirs.
		Note that this ignores the case where an AppDir turns into a 
		regular directory.  Hopefully, this should be pretty rare. */
	l = dir->dir_subdirs;
	while (l != NULL) {
		d = l->data;
		if (is_appdir(d->path)) {
			rem = l;
			l = g_list_next(l);
			dir->dir_subdirs = g_list_remove_link(dir->dir_subdirs, rem);
			menudir_destroy(rem->data);
			g_list_free_1(rem);
			dir->modtime = (time_t)0;
		} else l = g_list_next(l);
	}
	
	changed = menudir_has_changed(dir);
	if (changed) menudir_update_contents(dir);
	
	/* Update the subdirs in case any of them have changed. */
	l = dir->dir_subdirs;
	while (l != NULL) {
		menudir_update(l->data);
		/* Remove empty directories. */
		if (menudir_is_empty(l->data)) {
			rem = l;
			l = g_list_next(l);
			dir->dir_subdirs = g_list_remove_link(dir->dir_subdirs, rem);
			menudir_destroy(rem->data);
			g_list_free_1(rem);
		} else l = g_list_next(l);
	}
	
	if (changed) menudir_make_widgets(dir);
	dir->is_updating = FALSE;
}

gboolean menudir_is_empty(RoxMenuDir *dir) {
	g_return_val_if_fail(dir != NULL, TRUE);
	if (dir->dir_contents == NULL && dir->dir_subdirs == NULL) {
		return TRUE;
	} else {
		return FALSE;
	}
}

void menudir_set_grouping(RoxMenuDir *dir, RoxMenuGrouping mode, 
                          gboolean recursive) {
	g_return_if_fail(dir != NULL);
								  
	dir->dir_grouping = mode;
	if (recursive) {
		GList *l;
		l = dir->dir_subdirs;
		while (l != NULL) {
			menudir_set_grouping(l->data, mode, recursive);
			l = g_list_next(l);
		}
	}	
}	

void menudir_set_show_files(RoxMenuDir *dir, gboolean show, 
                            gboolean recursive) {
	g_return_if_fail(dir != NULL);
									
	if (dir->show_files != show) {
		printf("Should be updating\n");
		dir->show_files = show;
		dir->modtime = (time_t)0;  /* Force a menu update. */
	}
	if (recursive) {
		GList *l;
		l = dir->dir_subdirs;
		while (l != NULL) {
			menudir_set_show_files(l->data, show, recursive);
			l = g_list_next(l);
		}
	}
}
